import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/utils/description_text.dart';
import 'package:flutter/material.dart';

import '../default_scaffold_widget.dart';
import 'default_list_item.dart';

class ListScaffoldWidget<T> extends StatefulWidget {
  final Widget title;
  final List<T> list;
  final List<Widget> appBarButtons;
  final bool outsourceMenuButtons;
  final ISelectableListItem Function(BuildContext context, State state, T item)
      itemBuilder;
  final Comparable Function(T) sortProperty;
  final List<Comparable> Function(T) sortProperties;
  final bool reverseSort;
  final Comparable Function(T) groupProperty;
  final Widget Function(T) groupTitle;
  final List<Widget> Function(BuildContext context,
          ListScaffoldWidgetState<T> listScaffoldWidgetState)
      selectionMenuButtonsBuilder;
  final bool titleFullSpace;
  final bool itemsAreSelectable;
  final List<Widget> menuButtons;
  final Widget topWidget;
  final Widget noItemWidget;
  final bool showBackArrow;

  const ListScaffoldWidget({
    Key key,
    this.title,
    this.itemBuilder,
    this.list,
    this.sortProperty,
    this.groupProperty,
    this.appBarButtons,
    this.selectionMenuButtonsBuilder,
    List<Widget> selectionMenuButtons,
    this.outsourceMenuButtons,
    this.groupTitle,
    this.titleFullSpace = false,
    this.itemsAreSelectable = true,
    this.menuButtons,
    this.topWidget,
    this.showBackArrow = false,
    this.reverseSort = false,
    this.sortProperties,
    this.noItemWidget,
  })  : assert(list != null),
        assert(itemBuilder != null),
        assert(sortProperty == null || sortProperties == null),
        super(key: key);

  @override
  ListScaffoldWidgetState<T> createState() => ListScaffoldWidgetState<T>();
}

class ListScaffoldWidgetState<T> extends State<ListScaffoldWidget<T>> {
  bool selectionActive = false;
  List<T> selectedItems;

  List<List<T>> groupedItems;

  @override
  void initState() {
    super.initState();

    selectedItems = List<T>();
  }

  @override
  Widget build(BuildContext context) {
    _sortList();

    return DefaultScaffoldWidget(
      title: _getTitleWidget(),
      appBarButtons: selectionActive ? null : widget.appBarButtons,
      showBackArrow: widget.showBackArrow || selectionActive,
      outsourceMenuButtons: widget.outsourceMenuButtons,
      titleFullSpace: widget.titleFullSpace,
      onWillPop: () {
        if (selectionActive) {
          cancelSelection();
          return Future.value(false);
        } else
          return Future.value(true);
      },
      scrollable: false,
      child: widget.list.length == 0 ? _buildNoItemWidget() : _getListView(),
      menuButtons: selectionActive && widget.selectionMenuButtonsBuilder != null
          ? widget.selectionMenuButtonsBuilder(context, this)
          : widget.menuButtons,
    );
  }

  void _sortList() {
    if (widget.sortProperty != null)
      widget.list.sort((item1, item2) => widget
          .sortProperty(widget.reverseSort ? item2 : item1)
          .compareTo(widget.sortProperty(widget.reverseSort ? item1 : item2)));
    if (widget.sortProperties != null) {
      widget.list.sort((item1, item2) {
        var p1 = widget.sortProperties(widget.reverseSort ? item2 : item1);
        var p2 = widget.sortProperties(widget.reverseSort ? item1 : item2);
        for (var i = 0; i < p1.length; i++) {
          var result = p1[i].compareTo(p2[i]);
          if (result != 0) return result;
        }
        return 0;
      });
    }

    groupedItems = List<List<T>>();
    if (widget.groupProperty == null) {
      groupedItems.add(widget.list);
    } else if (widget.list.length > 0) {
      var lastList = List<T>();
      groupedItems.add(lastList);
      widget.list.asMap().forEach((i, item) {
        if (i == 0)
          lastList.add(item);
        else if (widget
                .groupProperty(lastList.last)
                .compareTo(widget.groupProperty(item)) ==
            0)
          lastList.add(item);
        else {
          lastList = List<T>();
          lastList.add(item);
          groupedItems.add(lastList);
        }
      });
    }
  }

  Widget _buildNoItemWidget() {
    return Column(
      children: <Widget>[
        widget.topWidget ?? Container(),
        DescriptionText.fromChild(widget.noItemWidget),
      ],
    );
  }

  Widget _getListView() {
    var offset = widget.topWidget == null ? 0 : 1;
    return ListView.builder(
      itemCount: groupedItems.length + offset,
      itemBuilder: (context, groupIndex) {
        if (widget.topWidget != null && groupIndex == 0)
          return widget.topWidget;

        return ItemSectionWidget.builder(
          title: widget.groupProperty == null
              ? null
              : widget.groupTitle == null
                  ? Text(widget
                      .groupProperty(groupedItems[groupIndex - offset].first)
                      .toString())
                  : widget.groupTitle(groupedItems[groupIndex - offset].first),
          itemCount: groupedItems[groupIndex - offset].length,
          itemBuilder: (context, itemIndex) {
            var item = groupedItems[groupIndex - offset][itemIndex];
            return widget.itemBuilder(context, this, item).copyWithSelection(
                isSelected: selectedItems.contains(item),
                onTapList: () {
                  if (selectionActive)
                    setState(() {
                      if (!selectedItems.remove(item)) selectedItems.add(item);
                    });
                  return !selectionActive;
                },
                onLongTapList: () {
                  if (widget.itemsAreSelectable)
                    setState(() {
                      selectionActive = true;
                      if (!selectedItems.remove(item)) selectedItems.add(item);
                    });
                });
          },
        );
      },
    );
  }

  Widget _getTitleWidget() {
    return selectionActive
        ? Text(
            selectedItems.length.toString() + " ausgewählt",
          )
        : widget.title;
  }

  void cancelSelection() {
    setState(() {
      selectionActive = false;
      selectedItems.clear();
    });
  }
}
