import 'dart:core';

import 'package:flutter/material.dart';

class Validator<T> {
  final List<IValidation> _validations = List<IValidation>();

  TextValidation createTextValidation(
      {@required String Function(T item, String text) validateFunction,
      FocusNode focusNode}) {
    var validation = TextValidation<T>(
      validateFunction: validateFunction,
      focusNode: focusNode,
    );
    _validations.add(validation);
    return validation;
  }

  DateTimeValidation createDateTimeValidation(
      {@required String Function(T item, DateTime dateTime) validateFunction,
      FocusNode focusNode}) {
    var validation = DateTimeValidation<T>(
      validateFunction: validateFunction,
      focusNode: focusNode,
    );
    _validations.add(validation);
    return validation;
  }

  bool isValid() {
    _validations.forEach((validation) => validation?.validate());
    return _validations.every((validation) => validation?.validate() ?? true);
  }

  FocusNode firstInvalid() {
    return _validations
        .firstWhere((validation) => !validation.validate(), orElse: null)
        ?.focusNode;
  }
}

abstract class IValidation {
  IValidation({this.focusNode});
  final FocusNode focusNode;

  bool validate();
}

class TextValidation<T> implements IValidation {
  final String Function(T, String string) validateFunction;
  final FocusNode focusNode;

  String Function() _getText;
  T Function() _getItem;
  void Function(String errorMessage) _onValidate;

  TextValidation({@required this.validateFunction, this.focusNode});

  @override
  bool validate() {
    if (_getItem == null || _getText == null || validateFunction == null)
      return true;
    var errorMessage = validateFunction(_getItem(), _getText());
    _onValidate?.call(errorMessage);
    return errorMessage == null;
  }

  void initialize(
      {@required String Function() getTextFunction,
      @required T Function() getItemFunction,
      @required void Function(String errorMessage) onValidate}) {
    _getText = getTextFunction;
    _getItem = getItemFunction;
    _onValidate = onValidate;
  }
}

class DateTimeValidation<T> implements IValidation {
  final String Function(T, DateTime dateTime) validateFunction;
  final FocusNode focusNode;

  DateTime Function() _getDateTime;
  T Function() _getItem;
  void Function(String errorMessage) _onValidate;

  DateTimeValidation({@required this.validateFunction, this.focusNode});

  @override
  bool validate() {
    if (_getItem == null || _getDateTime == null || validateFunction == null)
      return true;
    var errorMessage = validateFunction(_getItem(), _getDateTime());
    _onValidate?.call(errorMessage);
    return errorMessage == null;
  }

  void initialize(
      {@required DateTime Function() getDateTimeFunction,
      @required T Function() getItemFunction,
      @required void Function(String errorMessage) onValidate}) {
    _getDateTime = getDateTimeFunction;
    _getItem = getItemFunction;
    _onValidate = onValidate;
  }
}
